import unittest
import numpy as np
import graphepp as gg
from graphepp.graphepp import _mask_a, _mask_b, _mask_k


def random_graph(num_vertices, p=0.5):
    """Generate random gg.Graph.

    Graph with `num_vertices` vertices. Each edge exists with probability `p`.
    """
    edges = []
    for i in range(num_vertices):
        for j in range(i + 1, num_vertices):
            if np.random.random() < p:
                edges += [(i, j)]
    return gg.Graph(N=num_vertices, E=edges)


class TestGraph(unittest.TestCase):
    def test_immutableish(self):
        # test that all the properties are read only
        test_graph = gg.Graph(
            N=5, E=[(0, 1), (1, 2), (2, 0), (3, 4)], sets=[[0, 3], [1, 4], [2]]
        )
        with self.assertRaises(AttributeError):
            test_graph.N = 3
        with self.assertRaises(AttributeError):
            test_graph.E = ((1, 2),)
        with self.assertRaises(TypeError):
            E = test_graph.E
            E[0] = (0, 1)
        with self.assertRaises(AttributeError):
            test_graph.sets = ((1, 2), (0, 3))
        with self.assertRaises(TypeError):
            sets = test_graph.sets
            sets[0] = (0, 2, 4)
        with self.assertRaises(AttributeError):
            test_graph.adj = np.array([[0, 1], [1, 0]], dtype=int)
        adj = test_graph.adj
        adj[0, 0] = 1
        adj[1, 0] = 0
        self.assertFalse(np.all(adj == test_graph.adj))

    def test_equivalence(self):
        # test that equivalent graphs are considered equal
        test_graph = gg.Graph(N=5, E=[(0, 1), (0, 2), (2, 3), (3, 4)])
        # other number of vertices
        graph1 = gg.Graph(N=6, E=[(0, 1), (0, 2), (2, 3), (3, 4)])
        self.assertNotEqual(graph1, test_graph)
        self.assertNotEqual(hash(graph1), hash(test_graph))
        # order of nodes in edges
        graph2 = gg.Graph(N=5, E=[(1, 0), (0, 2), (3, 2), (4, 3)])
        self.assertEqual(graph2, test_graph)
        self.assertEqual(hash(graph2), hash(test_graph))
        # order of edges should not matter
        graph3 = gg.Graph(N=5, E=[(0, 1), (3, 4), (0, 2), (2, 3)])
        self.assertEqual(graph3, test_graph)
        self.assertEqual(hash(graph3), hash(test_graph))
        # different sets are not equivalent
        test_graph1 = gg.Graph(
            N=5, E=[(0, 1), (0, 2), (2, 3), (3, 4)], sets=[[0, 3], [1, 2, 4]]
        )
        self.assertNotEqual(test_graph1, test_graph)
        self.assertNotEqual(hash(test_graph1), hash(test_graph))
        graph4 = gg.Graph(
            N=5, E=[(0, 1), (0, 2), (2, 3), (3, 4)], sets=[[0], [1, 2, 4], [3]]
        )
        self.assertNotEqual(graph4, test_graph1)
        self.assertNotEqual(hash(graph4), hash(test_graph1))
        # order within sets should not matter
        graph5 = gg.Graph(
            N=5, E=[(0, 1), (0, 2), (2, 3), (3, 4)], sets=[[3, 0], [1, 4, 2]]
        )
        self.assertEqual(graph5, test_graph1)
        self.assertEqual(hash(graph5), hash(test_graph1))
        # but order of sets matters
        graph6 = gg.Graph(
            N=5, E=[(0, 1), (0, 2), (2, 3), (3, 4)], sets=[[1, 2, 4], [0, 3]]
        )
        self.assertNotEqual(graph6, test_graph1)
        self.assertNotEqual(hash(graph6), hash(test_graph1))


class TestNoise(unittest.TestCase):
    # the gg.noisy function does all the heavy lifting in this part,
    # the rest is basically just written down definitions
    """Tests for all the noise functions."""

    def test_noisy(self):
        # test small case with random numbers
        num_qubits = 2
        test_rho = np.random.random(2 ** num_qubits)
        test_rho = test_rho / np.sum(test_rho)
        test_result = gg.noisy(rho=test_rho, subset=[num_qubits - 1])
        known_result = np.array([test_rho[1], test_rho[0], test_rho[3], test_rho[2]])
        self.assertTrue(np.allclose(test_result, known_result))
        # test all single qubit patterns for an 8 qubit state
        num_qubits = 8
        test_rho = np.random.random(2 ** num_qubits)
        test_rho = test_rho / np.sum(test_rho)
        for qubit_index in range(num_qubits):
            test_result = gg.noisy(rho=test_rho, subset=[qubit_index])
            known_result = np.zeros_like(test_rho)
            binary_mask = 1 << ((num_qubits - 1) - qubit_index)
            for i in range(2 ** num_qubits):
                known_result[i] = test_rho[i ^ binary_mask]
            self.assertTrue(np.allclose(test_result, known_result))
        # test random patterns on states of random size
        for _ in range(256):
            num_qubits = np.random.randint(
                1, 16
            )  # up to 15 qubits shouldn't take too long to test
            test_rho = np.random.random(2 ** num_qubits)
            test_rho = test_rho / np.sum(test_rho)
            amount_choices = np.random.randint(num_qubits)
            chosen_qubits = np.sort(
                np.random.choice(num_qubits, size=amount_choices, replace=False)
            )
            test_result = gg.noisy(rho=test_rho, subset=chosen_qubits)
            binary_string = "0b"
            for i in range(num_qubits):
                if i in chosen_qubits:
                    binary_string += "1"
                else:
                    binary_string += "0"
            binary_mask = int(binary_string, 2)
            known_result = np.zeros_like(test_rho)
            for i in range(2 ** num_qubits):
                known_result[i] = test_rho[i ^ binary_mask]
            self.assertTrue(np.allclose(test_result, known_result))

    def test_znoisy(self):
        # kinda pointless because this is literally just the definition of
        # the function replicated
        num_qubits = 8
        test_rho = np.random.random(2 ** num_qubits)
        test_rho = test_rho / np.sum(test_rho)
        for i in range(num_qubits):
            test_result = gg.znoisy(rho=test_rho, qubit_index=i)
            # test calling with optional graph argument that does nothing
            test_result2 = gg.znoisy(
                rho=test_rho, qubit_index=i, graph=random_graph(num_qubits)
            )
            known_result = gg.noisy(rho=test_rho, subset=[i])
            self.assertTrue(np.allclose(test_result, known_result))
            self.assertTrue(np.allclose(test_result2, known_result))

    def test_xnoisy(self):
        # use an alternative way to do the same thing (but with more unnecessary function calls)
        num_qubits = 8
        test_rho = np.random.random(2 ** num_qubits)
        test_rho = test_rho / np.sum(test_rho)
        for i in range(num_qubits):
            test_graph = random_graph(num_qubits)
            test_result = gg.xnoisy(rho=test_rho, qubit_index=i, graph=test_graph)
            known_result = test_rho
            for j in range(
                num_qubits
            ):  # xnoise is equivalent to znoise on all neighbors
                if test_graph.adj[i, j]:
                    known_result = gg.znoisy(rho=known_result, qubit_index=j)
            self.assertTrue(np.allclose(test_result, known_result))

    def test_ynoisy(self):
        # use an alternative way to do the same thing (but with more unnecessary function calls)
        num_qubits = 8
        test_rho = np.random.random(2 ** num_qubits)
        test_rho = test_rho / np.sum(test_rho)
        for i in range(num_qubits):
            test_graph = random_graph(num_qubits)
            test_result = gg.ynoisy(rho=test_rho, qubit_index=i, graph=test_graph)
            # ynoise is equivalent to znoise on all neighbors AND the qubit itself
            known_result = gg.znoisy(rho=test_rho, qubit_index=i)
            for j in range(num_qubits):
                if test_graph.adj[i, j]:
                    known_result = gg.znoisy(rho=known_result, qubit_index=j)
            self.assertTrue(np.allclose(test_result, known_result))

    def test_znoise(self):
        # only reproducibility test because this function is just a definition
        test_rho = np.array([0.93, 0.01, 0.01, 0.01, 0.01, 0.01, 0.01, 0.01])
        test_graph = gg.Graph(N=3, E=[(0, 1), (0, 2)])  # GHZ graph
        test_result = gg.znoise(rho=test_rho, qubit_index=1, p=0.98, graph=test_graph)
        self.assertTrue(
            np.allclose(
                test_result,
                np.array([0.9116, 0.01, 0.0284, 0.01, 0.01, 0.01, 0.01, 0.01]),
            )
        )
        # now try different noise on all qubits
        ps = [0.7, 0.93, 0.98]
        test_result = test_rho
        for idx, p in enumerate(ps):
            test_result = gg.znoise(
                rho=test_result, qubit_index=idx, p=p, graph=test_graph
            )
        self.assertTrue(
            np.allclose(
                test_result,
                np.array(
                    [
                        0.5969416,
                        0.0219784,
                        0.0541784,
                        0.0109016,
                        0.2615464,
                        0.0151336,
                        0.0289336,
                        0.0103864,
                    ]
                ),
            )
        )

    def test_xnoise(self):
        # only reproducibility test because this function is just a definition
        test_rho = np.array([0.93, 0.01, 0.01, 0.01, 0.01, 0.01, 0.01, 0.01])
        test_graph = gg.Graph(N=3, E=[(0, 1), (0, 2)])  # GHZ graph
        test_result = gg.xnoise(rho=test_rho, qubit_index=1, p=0.98, graph=test_graph)
        self.assertTrue(
            np.allclose(
                test_result,
                np.array([0.9116, 0.01, 0.01, 0.01, 0.0284, 0.01, 0.01, 0.01]),
            )
        )
        # now try different noise on all qubits
        ps = [0.7, 0.93, 0.98]
        test_result = test_rho
        for idx, p in enumerate(ps):
            test_result = gg.xnoise(
                rho=test_result, qubit_index=idx, p=p, graph=test_graph
            )
        self.assertTrue(
            np.allclose(
                test_result,
                np.array(
                    [0.5978432, 0.01, 0.01, 0.2619328, 0.0661568, 0.01, 0.01, 0.0340672]
                ),
            )
        )

    def test_ynoise(self):
        # only reproducibility test because this function is just a definition
        test_rho = np.array([0.93, 0.01, 0.01, 0.01, 0.01, 0.01, 0.01, 0.01])
        test_graph = gg.Graph(N=3, E=[(0, 1), (0, 2)])  # GHZ graph
        test_result = gg.ynoise(rho=test_rho, qubit_index=1, p=0.98, graph=test_graph)
        self.assertTrue(
            np.allclose(
                test_result,
                np.array([0.9116, 0.01, 0.01, 0.01, 0.01, 0.01, 0.0284, 0.01]),
            )
        )
        # now try different noise on all qubits
        ps = [0.7, 0.93, 0.98]
        test_result = test_rho
        for idx, p in enumerate(ps):
            test_result = gg.ynoise(
                rho=test_result, qubit_index=idx, p=p, graph=test_graph
            )
        self.assertTrue(
            np.allclose(
                test_result,
                np.array(
                    [
                        0.5969416,
                        0.0289336,
                        0.0151336,
                        0.0109016,
                        0.0103864,
                        0.0219784,
                        0.0541784,
                        0.2615464,
                    ]
                ),
            )
        )

    def test_wnoise(self):
        # only reproducibility test because this function is just a definition
        test_rho = np.array([0.93, 0.01, 0.01, 0.01, 0.01, 0.01, 0.01, 0.01])
        test_graph = gg.Graph(N=3, E=[(0, 1), (0, 2)])  # GHZ graph
        test_result = gg.wnoise(rho=test_rho, qubit_index=1, p=0.98, graph=test_graph)
        self.assertTrue(
            np.allclose(
                test_result,
                np.array([0.9162, 0.01, 0.0146, 0.01, 0.0146, 0.01, 0.0146, 0.01]),
            )
        )
        # now try different noise on all qubits
        ps = [0.7, 0.93, 0.98]
        test_result = test_rho
        for idx, p in enumerate(ps):
            test_result = gg.wnoise(
                rho=test_result, qubit_index=idx, p=p, graph=test_graph
            )
        self.assertTrue(
            np.allclose(
                test_result,
                np.array(
                    [
                        0.6770368,
                        0.016164,
                        0.024214,
                        0.076056,
                        0.0900952,
                        0.016164,
                        0.024214,
                        0.076056,
                    ]
                ),
            )
        )

    def test_noise_pattern(self):
        # first compare to other functions
        test_rho = np.random.random(16)
        test_rho[0] += 16
        test_rho = test_rho / np.sum(test_rho)
        test_graph = random_graph(num_vertices=4, p=0.9)
        px = 0.5 + np.random.random() * 0.5
        known_result = gg.xnoise(rho=test_rho, qubit_index=2, p=px, graph=test_graph)
        test_result = gg.noise_pattern(
            rho=test_rho, qubit_index=2, ps=[px, 1 - px, 0, 0], graph=test_graph
        )
        self.assertTrue(np.allclose(test_result, known_result))
        py = 0.5 + np.random.random() * 0.5
        known_result = gg.ynoise(rho=test_rho, qubit_index=2, p=py, graph=test_graph)
        test_result = gg.noise_pattern(
            rho=test_rho, qubit_index=2, ps=[py, 0, 1 - py, 0], graph=test_graph
        )
        self.assertTrue(np.allclose(test_result, known_result))
        pz = 0.5 + np.random.random() * 0.5
        known_result = gg.znoise(rho=test_rho, qubit_index=2, p=pz, graph=test_graph)
        test_result = gg.noise_pattern(
            rho=test_rho, qubit_index=2, ps=[pz, 0, 0, 1 - pz], graph=test_graph
        )
        self.assertTrue(np.allclose(test_result, known_result))
        pw = 0.5 + np.random.random() * 0.5
        known_result = gg.wnoise(rho=test_rho, qubit_index=2, p=pw, graph=test_graph)
        test_result = gg.noise_pattern(
            rho=test_rho,
            qubit_index=2,
            ps=[pw + (1 - pw) / 4, (1 - pw) / 4, (1 - pw) / 4, (1 - pw) / 4],
            graph=test_graph,
        )
        self.assertTrue(np.allclose(test_result, known_result))
        # and reproducibility
        test_rho = np.array(
            [
                0.85,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
            ]
        )
        test_graph = gg.Graph(N=4, E=[(0, 1), (1, 2), (2, 3), (3, 0)])
        ps = [0.8, 0.1, 0.07, 0.03]
        test_result = gg.noise_pattern(
            rho=test_rho, qubit_index=2, ps=ps, graph=test_graph
        )
        self.assertTrue(
            np.allclose(
                test_result,
                np.array(
                    [
                        0.682,
                        0.01,
                        0.0352,
                        0.01,
                        0.01,
                        0.094,
                        0.01,
                        0.0688,
                        0.01,
                        0.01,
                        0.01,
                        0.01,
                        0.01,
                        0.01,
                        0.01,
                        0.01,
                    ]
                ),
            )
        )

    def test_wnoise_all(self):
        # again, this function is just writing down a definition
        test_rho = np.array([0.93, 0.01, 0.01, 0.01, 0.01, 0.01, 0.01, 0.01])
        test_graph = gg.Graph(N=3, E=[(0, 1), (0, 2)])  # GHZ graph
        p = 0.98
        test_result = gg.wnoise_all(rho=test_rho, p=p, graph=test_graph)
        self.assertTrue(
            np.allclose(
                test_result,
                np.array(
                    [
                        0.88928632,
                        0.014554,
                        0.014554,
                        0.014554,
                        0.02338968,
                        0.014554,
                        0.014554,
                        0.014554,
                    ]
                ),
            )
        )

    def test_noise_global(self):
        # again, this function is just following the definition
        test_rho = np.array([0.91, 0.02, 0.01, 0.01, 0.01, 0.01, 0.01, 0.02])
        test_graph = gg.Graph(N=3, E=[(0, 1), (0, 2)])  # GHZ graph
        p = 0.98
        test_result = gg.noise_global(rho=test_rho, p=p, graph=test_graph)
        self.assertTrue(
            np.allclose(
                test_result,
                np.array(
                    [0.8943, 0.0221, 0.0123, 0.0123, 0.0123, 0.0123, 0.0123, 0.0221]
                ),
            )
        )


class TestDistanceMeasures(unittest.TestCase):
    def _distance_measure_criterium(self, func):
        num_qubits = 5
        rho = np.random.random(2 ** num_qubits)
        rho = rho / np.sum(rho)
        # distance measure is 0 for same state
        self.assertAlmostEqual(func(rho, rho), 0)
        # distance measure is > 0 for different states
        sigma = np.random.random(2 ** num_qubits)
        sigma = sigma / np.sum(sigma)
        self.assertGreater(func(rho, sigma), 0)

    def test_fidelity(self):
        def dist_meas(rho, mu):
            return np.sqrt(np.abs(1 - gg.fidelity(rho, mu)))

        self._distance_measure_criterium(dist_meas)

        def dist_meas_2(rho, mu):
            return 1 - np.sqrt(gg.fidelity(rho, mu))

        self._distance_measure_criterium(dist_meas_2)
        # now a reproducibility test
        test_rho = np.array(
            [
                0.85,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
            ]
        )
        test_mu = np.array(
            [
                0.70,
                0.01,
                0.02,
                0.00,
                0.00,
                0.02,
                0.00,
                0.01,
                0.00,
                0.05,
                0.00,
                0.04,
                0.14,
                0.01,
                0.00,
                0.00,
            ]
        )
        self.assertAlmostEqual(gg.fidelity(test_rho, test_mu), 0.8270519315962255)

    def test_fid_alternative(self):
        def dist_meas(rho, mu):
            return np.sqrt(np.abs(1 - gg.fid_alternative(rho, mu) ** 2))

        self._distance_measure_criterium(dist_meas)

        def dist_meas_2(rho, mu):
            return 1 - gg.fid_alternative(rho, mu)

        self._distance_measure_criterium(dist_meas_2)
        test_rho = np.array(
            [
                0.85,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
            ]
        )
        test_mu = np.array(
            [
                0.70,
                0.01,
                0.02,
                0.00,
                0.00,
                0.02,
                0.00,
                0.01,
                0.00,
                0.05,
                0.00,
                0.04,
                0.14,
                0.01,
                0.00,
                0.00,
            ]
        )
        self.assertAlmostEqual(
            gg.fid_alternative(test_rho, test_mu), 0.9094239559172749
        )

    def test_trace_distance(self):
        self._distance_measure_criterium(gg.trace_distance)
        test_rho = np.array(
            [
                0.85,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
            ]
        )
        test_mu = np.array(
            [
                0.70,
                0.01,
                0.02,
                0.00,
                0.00,
                0.02,
                0.00,
                0.01,
                0.00,
                0.05,
                0.00,
                0.04,
                0.14,
                0.01,
                0.00,
                0.00,
            ]
        )
        self.assertAlmostEqual(gg.trace_distance(test_rho, test_mu), 0.22)

    def test_consistency(self):
        num_qubits = 5
        rho = np.random.random(2 ** num_qubits)
        rho = rho / np.sum(rho)
        sigma = np.random.random(2 ** num_qubits)
        sigma = sigma / np.sum(sigma)
        # assert relationship between fidelity and fid_alternative
        self.assertAlmostEqual(
            gg.fidelity(rho, sigma), gg.fid_alternative(rho, sigma) ** 2
        )
        # assert that the relation holds 1-sqrt(F) <= trace_distance <= sqrt(1-F)
        self.assertLessEqual(
            1 - np.sqrt(gg.fidelity(rho, sigma)), gg.trace_distance(rho, sigma)
        )
        self.assertLessEqual(
            gg.trace_distance(rho, sigma), np.sqrt(np.abs(1 - gg.fidelity(rho, sigma)))
        )


class TestAuxiliaryFunctions(unittest.TestCase):
    def test_normalize(self):
        num_qubits = 5
        rho = np.random.random(2 ** num_qubits)
        self.assertAlmostEqual(np.sum(gg.normalize(rho)), 1)
        # assert catching of numerical phenomena (usually happens when subtracting similar states)
        rho[0] = -1e-16
        rho[2] = -2.4e-15
        self.assertTrue(np.all(gg.normalize(rho) >= 0))
        # should never be called with large negative numbers, but make sure that norm is 1 anyway
        rho[2] = -2.4e-5
        self.assertAlmostEqual(np.sum(gg.normalize(rho)), 1)

    def test_local_complementation(self):
        # first test with known examples
        test_graph = gg.Graph(N=3, E=[(0, 1), (0, 2)])
        test_result = gg.local_complementation(n=0, graph=test_graph)
        known_result = gg.Graph(N=3, E=[(0, 1), (0, 2), (1, 2)])
        self.assertEqual(test_result, known_result)
        test_graph = gg.Graph(
            N=6, E=[(0, 1), (1, 2), (0, 3), (1, 4), (2, 5), (3, 4), (4, 5)]
        )  # butterfly graph
        test_result = gg.local_complementation(n=1, graph=test_graph)
        known_result = gg.Graph(
            N=6,
            E=[
                (0, 1),
                (1, 2),
                (0, 3),
                (1, 4),
                (2, 5),
                (3, 4),
                (4, 5),
                (0, 4),
                (2, 4),
                (0, 2),
            ],
        )
        self.assertEqual(test_result, known_result)
        # test with random graphs and stupid verification function
        for _ in range(10):
            num_vertices = 10
            test_graph = random_graph(num_vertices)
            for lc_vertex in range(num_vertices):  # try all possible complementations
                complemented_graph = gg.local_complementation(
                    n=lc_vertex, graph=test_graph
                )
                # now do the verification
                for i in range(num_vertices):
                    for j in range(i + 1, num_vertices):
                        if (
                            test_graph.adj[j, lc_vertex]
                            and test_graph.adj[i, lc_vertex]
                        ):  # i and j are both in the neighborhood of lc_vertex
                            self.assertEqual(
                                complemented_graph.adj[i, j],
                                (test_graph.adj[i, j] + 1) % 2,
                            )
                        else:
                            self.assertEqual(
                                complemented_graph.adj[i, j], test_graph.adj[i, j]
                            )


class TestTwoColorableEPP(unittest.TestCase):
    def test_mask_a(self):
        for _ in range(10):
            num_vertices = 10
            test_graph = random_graph(num_vertices)
            # randomly split in two sets - actual colorability doesn't matter for this test
            first_set = sorted(
                np.random.choice(
                    np.arange(num_vertices, dtype=int),
                    size=np.random.randint(num_vertices),
                    replace=False,
                )
            )
            second_set = [i for i in range(num_vertices) if i not in first_set]
            test_graph = gg.Graph(
                N=test_graph.N, E=test_graph.E, sets=[first_set, second_set]
            )
            for short_bit_string in range(2 ** len(test_graph.a)):
                long_bit_string = _mask_a(j=short_bit_string, graph=test_graph)
                short_string = format(
                    short_bit_string, "0" + str(len(test_graph.a)) + "b"
                )
                long_string = format(long_bit_string, "0" + str(test_graph.N) + "b")
                for bit, idx in zip(short_string, test_graph.a):
                    self.assertEqual(bit, long_string[idx])

    def test_mask_b(self):
        for _ in range(10):
            num_vertices = 10
            test_graph = random_graph(num_vertices)
            # randomly split in two sets - actual colorability doesn't matter for this test
            first_set = sorted(
                np.random.choice(
                    np.arange(num_vertices, dtype=int),
                    size=np.random.randint(num_vertices),
                    replace=False,
                )
            )
            second_set = [i for i in range(num_vertices) if i not in first_set]
            test_graph = gg.Graph(
                N=test_graph.N, E=test_graph.E, sets=[first_set, second_set]
            )
            for short_bit_string in range(2 ** len(test_graph.b)):
                long_bit_string = _mask_b(j=short_bit_string, graph=test_graph)
                short_string = format(
                    short_bit_string, "0" + str(len(test_graph.b)) + "b"
                )
                long_string = format(long_bit_string, "0" + str(test_graph.N) + "b")
                for bit, idx in zip(short_string, test_graph.b):
                    self.assertEqual(bit, long_string[idx])

    def test_p1(self):
        # only reproducibility tests for now - an alternative implementation with density matrices would be ideal
        test_graph = gg.Graph(
            N=4, E=((0, 1), (0, 2), (0, 3)), sets=[[0], [1, 2, 3]]
        )  # 4 qubit GHZ graph; two-colorable
        test_rho = np.array(
            [
                0.70,
                0.01,
                0.02,
                0.00,
                0.00,
                0.02,
                0.00,
                0.01,
                0.00,
                0.05,
                0.00,
                0.04,
                0.14,
                0.01,
                0.00,
                0.00,
            ]
        )
        test_result = gg.p1(rho=test_rho, graph=test_graph)
        self.assertTrue(
            np.allclose(
                test_result,
                np.array(
                    [
                        7.72984887e-01,
                        2.20403023e-02,
                        4.47103275e-02,
                        6.29722922e-04,
                        6.29722922e-04,
                        4.47103275e-02,
                        3.14861461e-04,
                        2.32997481e-02,
                        3.74685139e-02,
                        4.40806045e-03,
                        6.29722922e-03,
                        0.00000000e00,
                        1.57430730e-03,
                        2.20403023e-02,
                        1.25944584e-03,
                        1.76322418e-02,
                    ]
                ),
            )
        )
        test_graph = gg.Graph(
            N=5, E=((0, 1), (1, 2), (2, 3), (3, 4)), sets=[[0, 2, 4], [1, 3]]
        )  # 5 qubit linear cluster state
        test_rho = np.zeros(2 ** 5, dtype=np.float)
        test_rho[0] = 1.0
        test_rho = gg.wnoise_all(rho=test_rho, p=0.99, graph=test_graph)
        test_result = gg.p1(rho=test_rho, graph=test_graph)
        self.assertTrue(
            np.allclose(
                test_result,
                np.array(
                    [
                        9.75112509e-01,
                        1.25003027e-05,
                        9.87411374e-03,
                        1.25003027e-05,
                        1.25028117e-05,
                        1.25003027e-05,
                        3.76852011e-07,
                        1.25003027e-05,
                        9.87411374e-03,
                        2.51861704e-07,
                        5.01148266e-03,
                        2.51861704e-07,
                        3.76852011e-07,
                        2.51861704e-07,
                        1.25021931e-05,
                        2.51861704e-07,
                        1.25003027e-05,
                        6.43999920e-10,
                        2.51861704e-07,
                        6.43999920e-10,
                        1.25003027e-05,
                        6.43999920e-10,
                        2.51861704e-07,
                        6.43999920e-10,
                        1.25003027e-05,
                        6.43999920e-10,
                        2.51861704e-07,
                        6.43999920e-10,
                        1.25003027e-05,
                        6.43999920e-10,
                        2.51861704e-07,
                        6.43999920e-10,
                    ]
                ),
            )
        )

    def test_p2(self):
        test_graph = gg.Graph(
            N=4, E=((0, 1), (0, 2), (0, 3)), sets=[[0], [1, 2, 3]]
        )  # 4 qubit GHZ graph; two-colorable
        test_rho = np.array(
            [
                0.70,
                0.01,
                0.02,
                0.00,
                0.00,
                0.02,
                0.00,
                0.01,
                0.00,
                0.05,
                0.00,
                0.04,
                0.14,
                0.01,
                0.00,
                0.00,
            ]
        )
        test_result = gg.p2(rho=test_rho, graph=test_graph)
        self.assertTrue(
            np.allclose(
                test_result,
                np.array(
                    [
                        9.49244479e-01,
                        5.03680744e-03,
                        7.74893452e-04,
                        3.09957381e-03,
                        3.79697792e-02,
                        9.68616815e-04,
                        0.00000000e00,
                        1.93723363e-04,
                        0.00000000e00,
                        1.93723363e-03,
                        0.00000000e00,
                        0.00000000e00,
                        0.00000000e00,
                        7.74893452e-04,
                        0.00000000e00,
                        0.00000000e00,
                    ]
                ),
            )
        )
        test_graph = gg.Graph(
            N=5, E=((0, 1), (1, 2), (2, 3), (3, 4)), sets=[[0, 2, 4], [1, 3]]
        )  # 5 qubit linear cluster state
        test_rho = np.zeros(2 ** 5, dtype=np.float)
        test_rho[0] = 1.0
        test_rho = gg.wnoise_all(rho=test_rho, p=0.99, graph=test_graph)
        test_result = gg.p2(rho=test_rho, graph=test_graph)
        self.assertTrue(
            np.allclose(
                test_result,
                np.array(
                    [
                        9.74888853e-01,
                        4.94873525e-03,
                        3.73657691e-05,
                        2.51192019e-05,
                        4.98572109e-03,
                        4.94873525e-03,
                        1.28738842e-05,
                        2.51192019e-05,
                        3.73657691e-05,
                        3.79924263e-07,
                        1.25002804e-05,
                        2.53094747e-07,
                        1.28738842e-05,
                        3.79924263e-07,
                        1.24996619e-05,
                        2.53094747e-07,
                        4.94873525e-03,
                        5.01102580e-05,
                        3.79924263e-07,
                        3.79299534e-07,
                        4.94873525e-03,
                        5.01102580e-05,
                        3.79924263e-07,
                        3.79299534e-07,
                        2.51192019e-05,
                        3.79299534e-07,
                        2.53094747e-07,
                        1.29395220e-07,
                        2.51192019e-05,
                        3.79299534e-07,
                        2.53094747e-07,
                        1.29395220e-07,
                    ]
                ),
            )
        )

    def test_p1_var(self):
        # assert that calling it with rho=sigma is equivalent to p1
        test_graph = gg.Graph(
            N=4, E=((0, 1), (0, 2), (0, 3)), sets=[[0], [1, 2, 3]]
        )  # 4 qubit GHZ graph; two-colorable
        test_rho = np.random.random(2 ** 4)
        test_rho = test_rho / np.sum(test_rho)
        test_result1 = gg.p1(rho=test_rho, graph=test_graph)
        test_result2 = gg.p1_var(rho=test_rho, sigma=test_rho, graph=test_graph)
        self.assertTrue(np.allclose(test_result1, test_result2))
        # also small reproducibility test
        test_rho = np.array(
            [
                0.70,
                0.01,
                0.02,
                0.00,
                0.00,
                0.02,
                0.00,
                0.01,
                0.00,
                0.05,
                0.00,
                0.04,
                0.14,
                0.01,
                0.00,
                0.00,
            ]
        )
        test_sigma = np.array(
            [
                0.84,
                0.01,
                0.00,
                0.01,
                0.00,
                0.02,
                0.00,
                0.01,
                0.00,
                0.00,
                0.05,
                0.02,
                0.03,
                0.01,
                0.00,
                0.00,
            ]
        )
        test_result = gg.p1_var(rho=test_rho, sigma=test_sigma, graph=test_graph)
        self.assertTrue(
            np.allclose(
                test_result,
                np.array(
                    [
                        8.37507114e-01,
                        2.21969266e-02,
                        2.46158224e-02,
                        1.02447353e-02,
                        7.11439954e-04,
                        4.41092772e-02,
                        5.69151964e-04,
                        2.24815026e-02,
                        7.25668754e-03,
                        5.26465566e-03,
                        1.42287991e-03,
                        3.55719977e-03,
                        7.11439954e-04,
                        2.13431986e-03,
                        1.08138873e-02,
                        6.40295959e-03,
                    ]
                ),
            )
        )

    def test_p2_var(self):
        # assert that calling it with rho=sigma is equivalent to p2
        test_graph = gg.Graph(
            N=4, E=((0, 1), (0, 2), (0, 3)), sets=[[0], [1, 2, 3]]
        )  # 4 qubit GHZ graph; two-colorable
        test_rho = np.random.random(2 ** 4)
        test_rho = test_rho / np.sum(test_rho)
        test_result1 = gg.p2(rho=test_rho, graph=test_graph)
        test_result2 = gg.p2_var(rho=test_rho, sigma=test_rho, graph=test_graph)
        self.assertTrue(np.allclose(test_result1, test_result2))
        # also small reproducibility test
        test_rho = np.array(
            [
                0.70,
                0.01,
                0.02,
                0.00,
                0.00,
                0.02,
                0.00,
                0.01,
                0.00,
                0.05,
                0.00,
                0.04,
                0.14,
                0.01,
                0.00,
                0.00,
            ]
        )
        test_sigma = np.array(
            [
                0.84,
                0.01,
                0.00,
                0.01,
                0.00,
                0.02,
                0.00,
                0.01,
                0.00,
                0.00,
                0.05,
                0.02,
                0.03,
                0.01,
                0.00,
                0.00,
            ]
        )
        test_result = gg.p2_var(rho=test_rho, sigma=test_sigma, graph=test_graph)
        self.assertTrue(
            np.allclose(
                test_result,
                np.array(
                    [
                        9.86577181e-01,
                        1.67785235e-04,
                        0.00000000e00,
                        1.34228188e-03,
                        7.04697987e-03,
                        8.38926174e-04,
                        0.00000000e00,
                        1.67785235e-04,
                        0.00000000e00,
                        8.38926174e-04,
                        1.67785235e-03,
                        6.71140940e-04,
                        0.00000000e00,
                        6.71140940e-04,
                        0.00000000e00,
                        0.00000000e00,
                    ]
                ),
            )
        )


class TestArbitraryGraphEPP(unittest.TestCase):
    def test_mask_k(self):
        for _ in range(10):
            num_vertices = 10
            test_graph = random_graph(num_vertices)
            # randomly pick a subset - actual colorability doesn't matter for this test
            random_set = tuple(
                sorted(
                    np.random.choice(
                        np.arange(num_vertices, dtype=int),
                        size=np.random.randint(num_vertices),
                        replace=False,
                    )
                )
            )
            for short_bit_string in range(2 ** len(random_set)):
                long_bit_string = _mask_k(
                    j=short_bit_string, graph=test_graph, subset=random_set
                )
                short_string = format(
                    short_bit_string, "0" + str(len(random_set)) + "b"
                )
                long_string = format(long_bit_string, "0" + str(test_graph.N) + "b")
                for bit, idx in zip(short_string, random_set):
                    self.assertEqual(bit, long_string[idx])

    def test_pk(self):
        # assert that for a two-colorables state with graph1 == graph2 and
        # appropriate subset, this is equivalent to p1_var or p2_var
        test_graph = gg.Graph(
            N=4, E=((0, 1), (0, 2), (0, 3)), sets=[[0], [1, 2, 3]]
        )  # 4 qubit GHZ graph; two-colorable
        for _ in range(10):
            test_rho = np.random.random(2 ** 4)
            test_rho = test_rho / np.sum(test_rho)
            test_sigma = np.random.random(2 ** 4)
            test_sigma = test_sigma / np.sum(test_sigma)
            test_result1 = gg.pk(
                rho=test_rho,
                sigma=test_sigma,
                graph1=test_graph,
                graph2=test_graph,
                subset=test_graph.a,
            )
            test_result2 = gg.pk(
                rho=test_rho,
                sigma=test_sigma,
                graph1=test_graph,
                graph2=test_graph,
                subset=test_graph.b,
            )
            known_result1 = gg.p1_var(rho=test_rho, sigma=test_sigma, graph=test_graph)
            known_result2 = gg.p2_var(rho=test_rho, sigma=test_sigma, graph=test_graph)
            self.assertTrue(np.allclose(test_result1, known_result1))
            self.assertTrue(np.allclose(test_result2, known_result2))
        # small reproducibility test
        test_graph = gg.Graph(
            N=4, E=((0, 1), (1, 2), (2, 0), (2, 3)), sets=[[0, 3], [1], [2]]
        )  # small graph that is actually not two-colorable, we choose a coloring with 3 colors
        subgraph0 = gg.Graph(N=4, E=((0, 1), (2, 0), (2, 3)), sets=[[0, 3], [1, 2]])
        subgraph1 = gg.Graph(N=4, E=((0, 1), (1, 2)), sets=[[1], [0, 2, 3]])
        subgraph2 = gg.Graph(N=4, E=((1, 2), (2, 0), (2, 3)), sets=[[2], [0, 1, 3]])
        test_rho = np.array(
            [
                0.70,
                0.01,
                0.02,
                0.00,
                0.00,
                0.02,
                0.00,
                0.01,
                0.00,
                0.05,
                0.00,
                0.04,
                0.14,
                0.01,
                0.00,
                0.00,
            ]
        )
        test_sigma = np.array(
            [
                0.84,
                0.01,
                0.00,
                0.01,
                0.00,
                0.02,
                0.00,
                0.01,
                0.00,
                0.00,
                0.05,
                0.02,
                0.03,
                0.01,
                0.00,
                0.00,
            ]
        )
        test_result0 = gg.pk(
            rho=test_rho,
            sigma=test_sigma,
            graph1=test_graph,
            graph2=subgraph0,
            subset=test_graph.sets[0],
        )
        self.assertTrue(
            np.allclose(
                test_result0,
                np.array(
                    [
                        9.46859903e-01,
                        9.66183575e-04,
                        2.70531401e-02,
                        8.05152979e-04,
                        0.00000000e00,
                        8.05152979e-04,
                        0.00000000e00,
                        6.44122383e-04,
                        6.76328502e-03,
                        1.44927536e-03,
                        0.00000000e00,
                        1.61030596e-03,
                        0.00000000e00,
                        8.05152979e-04,
                        1.12721417e-02,
                        9.66183575e-04,
                    ]
                ),
            )
        )
        test_result1 = gg.pk(
            rho=test_rho,
            sigma=test_sigma,
            graph1=test_graph,
            graph2=subgraph1,
            subset=test_graph.sets[1],
        )
        self.assertTrue(
            np.allclose(
                test_result1,
                np.array(
                    [
                        7.59674923e-01,
                        2.27038184e-02,
                        2.30908153e-02,
                        1.25128999e-02,
                        6.19195046e-03,
                        2.19298246e-03,
                        5.15995872e-04,
                        0.00000000e00,
                        2.45098039e-03,
                        5.46955624e-02,
                        4.65686275e-02,
                        6.20485036e-02,
                        5.15995872e-04,
                        4.38596491e-03,
                        2.57997936e-04,
                        2.19298246e-03,
                    ]
                ),
            )
        )
        test_result2 = gg.pk(
            rho=test_rho,
            sigma=test_sigma,
            graph1=test_graph,
            graph2=subgraph2,
            subset=test_graph.sets[2],
        )
        self.assertTrue(
            np.allclose(
                test_result2,
                np.array(
                    [
                        6.95285011e-01,
                        2.00562984e-02,
                        1.05559465e-03,
                        2.58034248e-03,
                        1.05559465e-03,
                        3.78841192e-02,
                        1.17288295e-04,
                        2.34576589e-04,
                        1.05559465e-03,
                        5.32488858e-02,
                        1.64203612e-03,
                        4.69153179e-04,
                        1.63969036e-01,
                        2.00562984e-02,
                        7.03729768e-04,
                        5.86441473e-04,
                    ]
                ),
            )
        )


if __name__ == "__main__":
    unittest.main()



"""Find the fixed point of an EPP with noisy CNOT gates."""

import os, sys

sys.path.insert(0, os.path.abspath(".."))
import graphepp as gg
import numpy as np


# CNOT gate is modelled as local white noise acting on both of the qubits
# followed by the perfect operation
noise_param = 0.98
ghz5_graph = gg.Graph(
    N=5, E=((0, i) for i in range(1, 5)), sets=[(0,), (i for i in range(1, 5))]
)  # 5-qubit GHZ state

print("Noisy fixed point of the P1-P2 protocol for the 5-qubit GHZ state.")
print(
    f"CNOTs are noisy with local depolarizing noise with error parameter p={noise_param:.2f}"
)

rho = np.zeros(2 ** ghz5_graph.N, dtype=np.float)
rho[0] = 1.0  # perfect state

for i in range(
    100
):  # 100 iterations are more than enough to reach the fixed point with numerical precision
    rho = gg.wnoise_all(
        rho=rho, p=noise_param, graph=ghz5_graph
    )  # a noisy CNOT is acting on each of the qubits to perform the protocol
    rho = gg.p1(rho=rho, graph=ghz5_graph)
    rho = gg.wnoise_all(rho=rho, p=noise_param, graph=ghz5_graph)
    rho = gg.p2(rho=rho, graph=ghz5_graph)
    fixed_point_fidelity_p2 = rho[0]  # fidelity when ending with p2
rho = gg.wnoise_all(
    rho=rho, p=noise_param, graph=ghz5_graph
)  # a noisy CNOT is acting on each of the qubits to perform the protocol
rho = gg.p1(rho=rho, graph=ghz5_graph)
fixed_point_fidelity_p1 = rho[0]  # fidelity when ending with p1
print(f"Fixed point fidelity when ending with P1: {fixed_point_fidelity_p1}")
print(f"Fixed point fidelity when ending with P2: {fixed_point_fidelity_p2}")

print("=================")
print("Fixed point fidelity as function of CNOT noise parameter")


def noisy_fixed_point(graph, noise_param):
    rho = np.zeros(2 ** graph.N, dtype=np.float)
    rho[0] = 1.0  # perfect state
    for i in range(
        100
    ):  # 100 iterations are more than enough to reach the fixed point with numerical precision
        rho = gg.wnoise_all(
            rho=rho, p=noise_param, graph=graph
        )  # a noisy CNOT is acting on each of the qubits to perform the protocol
        rho = gg.p1(rho=rho, graph=ghz5_graph)
        rho = gg.wnoise_all(rho=rho, p=noise_param, graph=graph)
        rho = gg.p2(rho=rho, graph=ghz5_graph)
    return rho


ghz5_graph = gg.Graph(
    N=5, E=((0, i) for i in range(1, 5)), sets=[(0,), (i for i in range(1, 5))]
)
noise_list = np.linspace(0.9, 1, num=21)
print("Calculating fidelity list... (this may take a bit)")
fidelity_list = [
    noisy_fixed_point(graph=ghz5_graph, noise_param=p)[0] for p in noise_list
]
print(f"{noise_list=}")
print(f"{fidelity_list=}")
print("One can see that below a certain noise threshold, the EPP stops working.")
print("If pyplot is available, will show a plot.")

try:
    import matplotlib.pyplot as plt

    plotting_possible = True
except ImportError:
    plotting_possible = False

if plotting_possible:
    plt.scatter(noise_list, fidelity_list)
    plt.title("5-qubit GHZ state at EPP fixed point")
    plt.xlabel("Error parameter p for CNOT gates")
    plt.ylabel("Fidelity at fixed point")
    plt.show()



"""Perform the entanglement purification protocol for arbitrary graph states.

Here we look at an example for a 6-qubit graph state that is three-colorable.
"""

import os, sys

sys.path.insert(0, os.path.abspath(".."))
import graphepp as gg
import numpy as np

# as main graph we use this 6-qubit state that is linked to a
# universal quantum error correction code
# For visualization of this particular coloring and the auxiliary graphs
# see Fig.9 and Fig. 20 in https://arxiv.org/abs/1609.05754
main_graph = gg.Graph(
    N=6,
    E=[(0, 1), (0, 4), (0, 5), (1, 2), (1, 3), (2, 3), (2, 5), (3, 4), (4, 5)],
    sets=[[0, 2], [1, 4], [3, 5]],
)
aux_graph1 = gg.Graph(N=6, E=[(0, 1), (0, 4), (0, 5), (1, 2), (2, 3), (2, 5)])
aux_graph2 = gg.Graph(N=6, E=[(0, 1), (0, 4), (1, 2), (1, 3), (3, 4), (4, 5)])
aux_graph3 = gg.Graph(N=6, E=[(0, 5), (1, 3), (2, 3), (2, 5), (3, 4), (4, 5)])

# assume we have all these states available, with the same noise per qubit
rho = np.zeros(2 ** main_graph.N, dtype=np.float)
rho[0] = 1.0  # perfect graph state
rho = gg.wnoise_all(rho=rho, p=0.98, graph=main_graph)

mu1 = np.zeros(2 ** aux_graph1.N, dtype=np.float)
mu1[0] = 1.0
mu1 = gg.wnoise_all(rho=mu1, p=0.98, graph=aux_graph1)

mu2 = np.zeros(2 ** aux_graph2.N, dtype=np.float)
mu2[0] = 1.0
mu2 = gg.wnoise_all(rho=mu2, p=0.98, graph=aux_graph2)

mu3 = np.zeros(2 ** aux_graph3.N, dtype=np.float)
mu3[0] = 1.0
mu3 = gg.wnoise_all(rho=mu3, p=0.98, graph=aux_graph3)

# do one protocol for each color, here we assume perfect CNOTs
print(f"Fidelity before: {rho[0]}")
rho = gg.pk(
    rho=rho, sigma=mu1, graph1=main_graph, graph2=aux_graph1, subset=main_graph.sets[0]
)
print(f"Fidelity after first subprotocol: {rho[0]:.5f}")
rho = gg.pk(
    rho=rho, sigma=mu2, graph1=main_graph, graph2=aux_graph2, subset=main_graph.sets[1]
)
print(f"Fidelity after second subprotocol: {rho[0]:.5f}")
rho = gg.pk(
    rho=rho, sigma=mu3, graph1=main_graph, graph2=aux_graph3, subset=main_graph.sets[2]
)
print(f"Fidelity after third subprotocol: {rho[0]:.5f}")



# -*- coding: utf-8 -*-
"""Functions for multiparite entanglement purification on graph states.

See README.md for an overview of the functionality.
"""
# All states are assumed to be diagonal in the graph state basis corresponding to
# the graph state considered.
# The noise should work for arbitrary graph states, while the purification
# protocol only makes sense for two-colorable graph states.
#
# Make sure the variables input follow the conventions given in the docstrings,
# since not many sanity checks are included.
#
# This should run reasonably well even for bigger states (ca. 10 qubits) if
# cythonized.

import numpy as np
from itertools import product
from functools import lru_cache


# ====Graph definitions==== #


def adj_matrix(N, E):
    """Calculates the adjacency matrix from vetices and edges.

    The graph has `N` vertices (labeled 0 to N-1) and edges `E`.

    Parameters
    ----------
    N : int
        Number of vertices.
    E : list (or tuple) of tuples
        Should contain 2-tuples with the edges of the graph. Each pair
        (i,j) indicates a connection between vertices i and j. Only
        simple, unweighted, undirected graphs are supported. Note that
        the `N` vertices are labeled 0...N-1

    Returns
    -------
    adj : np.ndarray
        Adjacency matrix of the graph specified. Is a symmetric `N` x `N` matrix
        with N_{ij}=1 if (i,j) is in `E` and 0 otherwise.

    """
    adj = np.zeros((N, N), dtype=int)
    for i, n in product(range(N), repeat=2):
        if (i, n) in E:
            adj[i, n] = 1
            adj[n, i] = 1
    return adj


class Graph(object):
    """A graph object consisting of vertices and edges.

    Other functions that need to know in which graph state basis a state is
    given expect a `Graph` object to specify the associated graph.
    The properties are read-only on purpose to make this hashable. This is
    desirable because some functions in this module profit heavily from caching.

    Parameters
    ----------
    N : int
        Number of vertices.
    E : list of tuples of ints
        Should contain 2-tuples with the edges of the graph. Each pair
        (i,j) indicates a connection between vertices i and j. Only
        simple, unweighted, undirected graphs are supported. Note that
        the `N` vertices are labeled 0...N-1
    sets : list of list of ints, optional
        Optionally define subsets of vertices, e.g. coloring of the graph
        as expected for the entanglement purification protocols. Default: []

    Attributes
    ----------
    adj : np.ndarray
        Adjacency matrix of the graph.
    N
    E
    sets
    a : tuple of ints
        the first subset of vertices (e.g. one color)
    b : tuple of ints
        the second subset of vertices (e.g. one color)

    """

    def __init__(self, N, E, sets=[]):
        self._N = N
        self._E = tuple(sorted(tuple(sorted(edge)) for edge in E))
        self._sets = tuple(tuple(sorted(set)) for set in sets)
        self._adj = tuple(tuple(row) for row in adj_matrix(N, E))

    def __repr__(self):
        return f"gg.Graph(N={self.N}, E={self.E}, sets={self.sets})"

    def __hash__(self):
        return hash((self._N, self._sets, self._adj))

    def __eq__(self, other):
        if isinstance(other, Graph):
            return (
                self.N == other.N
                and self.sets == other.sets
                and np.all(self.adj == other.adj)
            )
        else:
            return NotImplemented

    @property
    def N(self):
        return self._N

    @property
    def E(self):
        return self._E

    @property
    def sets(self):
        return self._sets

    @property
    def adj(self):
        return np.array(self._adj, dtype=int)

    @property
    def a(self):
        # subset a is the first color
        try:
            return self.sets[0]
        except IndexError:
            return None

    @property
    def b(self):
        # subset b is the first color
        try:
            return self.sets[1]
        except IndexError:
            return None


# ====Noise functions==== #


def noisy(rho, subset, graph=None):
    """Template to generate noise patterns.

    In physical terms this is correlated sigma_z noise on all particles in `subset`.

    Parameters
    ----------
    rho : np.ndarray
        Is the state acted on. Should be a 2**N-dimensional vector
        with the diagonal entries of the density matrix in the graph
        state basis.
    subset : list of int
        The list of which qubits are affected, counting starts at 0.
        Indices are expected to be in order
        012...(N-1) regardless of coloring of the vertices.
    graph : Graph, optional
        Specifies in which graph state basis rho is given.
        This function does not use it - only included for consistency. Default: None

    Returns
    -------
    np.ndarray
        The state after the action. Same shape as `rho`.
    """
    N = int(np.log2(len(rho)))
    rho = rho.reshape((2,) * N)
    rho = np.flip(rho, axis=subset)
    rho = rho.reshape((2 ** N,))
    return rho


#     # if there is only one int in subset, the following is actually faster
#     # than np.flip for some reason
#     for n in subset:
#         rho = np.swapaxes(np.swapaxes(rho, 0, n)[::-1], 0, n)

#    #old, slow implementation
#    j=0
#    for n in nn:
#        k=int(np.log2(len(rho)))-1-n
#        j=j^(1<<k) # + would be slightly faster than ^ but can lead to weird behaviour
#    mu=np.zeros(len(rho))
#    for i in range(len(mu)):
#        mu[i]=rho[i^j]
#    return mu


def znoisy(rho, qubit_index, graph=None):
    """Applies sigma_z noise on the specified qubit.

    Parameters
    ----------
    rho : np.ndarray
        Is the state acted on. Should be a 2**N-dimensional vector
        with the diagonal entries of the density matrix in the graph
        state basis.
    qubit_index : int
        The `qubit_index`-th qubit is affected, counting starts at 0. Indices
        are expected to be in order 012...(N-1) regardless of coloring of
        the vertices.
    graph : Graph, optional
        Specifies in which graph state basis rho is given.
        This function does not use it. Included only so znoisy can be called
        the same way as xnoisy and ynoisy. Default: None

    Returns
    -------
    np.ndarray
        The state after the action. Same shape as `rho`.
    """
    return noisy(rho, [qubit_index])


def xnoisy(rho, qubit_index, graph):
    """Applies sigma_x noise on the specified qubit.

    Parameters
    ----------
    rho : np.ndarray
        Is the state acted on. Should be a 2**N-dimensional vector
        with the diagonal entries of the density matrix in the graph
        state basis.
    qubit_index : int
        The `qubit_index`-th qubit is affected, counting starts at 0. Indices
        are expected to be in order 012...(N-1) regardless of coloring of
        the vertices.
    graph : Graph
        Specifies in which graph state basis rho is given.

    Returns
    -------
    np.ndarray
        The state after the action. Same shape as `rho`.
    """
    nn = []
    for i in range(graph.N):
        if graph.adj[qubit_index, i]:
            nn += [i]
    return noisy(rho, nn)


def ynoisy(rho, qubit_index, graph):
    """Applies sigma_y noise on the specified qubit.

    Parameters
    ----------
    rho : np.ndarray
        Is the state acted on. Should be a 2**N-dimensional vector
        with the diagonal entries of the density matrix in the graph
        state basis.
    qubit_index : int
        The `qubit_index`-th qubit is affected, counting starts at 0. Indices
        are expected to be in order 012...(N-1) regardless of coloring of
        the vertices.
    graph : Graph
        Specifies in which graph state basis rho is given.

    Returns
    -------
    np.ndarray
        The state after the action. Same shape as `rho`.
    """
    nn = [qubit_index]
    for i in range(graph.N):
        if graph.adj[qubit_index, i]:
            nn += [i]
    return noisy(rho, nn)


def znoise(rho, qubit_index, p, graph=None):
    """Apply Pauli-Z noise channel with error parameter `p` on a qubit.

    Parameters
    ----------
    rho : np.ndarray
        Is the state acted on. Should be a 2**N-dimensional vector
        with the diagonal entries of the density matrix in the graph
        state basis.
    qubit_index : int
        The `qubit_index`-th qubit is affected, counting starts at 0. Indices
        are expected to be in order 012...(N-1) regardless of coloring of
        the vertices.
    p : scalar
        Error parameter of the channel should be in interval [0, 1].
    graph : Graph, optional
        Specifies in which graph state basis rho is given. Default: None

    Returns
    -------
    np.ndarray
        The state after the action. Same shape as `rho`.

    """
    return p * rho + (1 - p) * znoisy(rho, qubit_index, graph)


def xnoise(rho, qubit_index, p, graph):
    """Apply Pauli-X noise channel with error parameter `p` on a qubit.

    Parameters
    ----------
    rho : np.ndarray
        Is the state acted on. Should be a 2**N-dimensional vector
        with the diagonal entries of the density matrix in the graph
        state basis.
    qubit_index : int
        The `qubit_index`-th qubit is affected, counting starts at 0. Indices
        are expected to be in order 012...(N-1) regardless of coloring of
        the vertices.
    p : scalar
        Error parameter of the channel should be in interval [0, 1].
    graph : Graph
        Specifies in which graph state basis rho is given.

    Returns
    -------
    np.ndarray
        The state after the action. Same shape as `rho`.

    """
    return p * rho + (1 - p) * xnoisy(rho, qubit_index, graph)


def ynoise(rho, qubit_index, p, graph):
    """Apply Pauli-Y noise channel with error parameter `p` on a qubit.

    Parameters
    ----------
    rho : np.ndarray
        Is the state acted on. Should be a 2**N-dimensional vector
        with the diagonal entries of the density matrix in the graph
        state basis.
    qubit_index : int
        The `qubit_index`-th qubit is affected, counting starts at 0. Indices
        are expected to be in order 012...(N-1) regardless of coloring of
        the vertices.
    p : scalar
        Error parameter of the channel should be in interval [0, 1].
    graph : Graph
        Specifies in which graph state basis rho is given.

    Returns
    -------
    np.ndarray
        The state after the action. Same shape as `rho`.

    """
    return p * rho + (1 - p) * ynoisy(rho, qubit_index, graph)


def wnoise(rho, qubit_index, p, graph):
    """Apply local white noise channel with error parameter `p` on a qubit.

    Note: local white noise is often also called local depolarizing noise

    Parameters
    ----------
    rho : np.ndarray
        Is the state acted on. Should be a 2**N-dimensional vector
        with the diagonal entries of the density matrix in the graph
        state basis.
    qubit_index : int
        The `qubit_index`-th qubit is affected, counting starts at 0. Indices
        are expected to be in order 012...(N-1) regardless of coloring of
        the vertices.
    p : scalar
        Error parameter of the channel should be in interval [0, 1].
    graph : Graph
        Specifies in which graph state basis rho is given.

    Returns
    -------
    np.ndarray
        The state after the action. Same shape as `rho`.

    """
    return p * rho + (1 - p) / 4 * (
        rho
        + xnoisy(rho, qubit_index, graph)
        + ynoisy(rho, qubit_index, graph)
        + znoisy(rho, qubit_index, graph)
    )


def noise_pattern(rho, qubit_index, ps, graph):
    """Applies a local pauli-diagonal noise channel on the specified qubit.

    Parameters
    ----------
    rho : np.ndarray
        Is the state acted on. Should be a 2**N-dimensional vector
        with the diagonal entries of the density matrix in the graph
        state basis.
    qubit_index : int
        The `n`-th qubit is affected, counting starts at 0. Indices are
        expected to be in order
        012...(N-1) regardless of coloring of the vertices.
    ps : list of scalars
        The coefficients of the noise channel.
        Should have 4 entries p_0 p_x p_y p_z that sum to 1.
    graph : Graph
        Specifies in which graph state basis rho is given.

    Returns
    -------
    np.ndarray
        The state after the action. Same shape as `rho`.
    """
    return (
        ps[0] * rho
        + ps[1] * xnoisy(rho, qubit_index, graph)
        + ps[2] * ynoisy(rho, qubit_index, graph)
        + ps[3] * znoisy(rho, qubit_index, graph)
    )


def wnoise_all(rho, p, graph):
    """Apply local white noise with the same error parameter to all qubits.

    Parameters
    ----------
    rho : np.ndarray
        Is the state acted on. Should be a 2**N-dimensional vector
        with the diagonal entries of the density matrix in the graph
        state basis.
    p : scalar
        Error parameter of the channel should be in interval [0, 1].
    graph : Graph
        Specifies the graphstate considered.

    Returns
    -------
    np.ndarray
        The state after the action. Same shape as `rho`.

    """
    for i in range(int(np.log2(len(rho)))):
        rho = wnoise(rho, i, p, graph)
    return rho


def noise_global(rho, p, graph=None):
    """Apply a global white noise channel to the state.

    Parameters
    ----------
    rho : np.ndarray
        Is the state acted on. Should be a 2**N-dimensional vector
        with the diagonal entries of the density matrix in the graph
        state basis.
    p : scalar
        Error parameter of the channel should be in interval [0, 1].
    graph : Graph, optional
        Specifies in which graph state basis rho is given.
        This function does not use it - only included for consistency. Default: None

    Returns
    -------
    np.ndarray
        The state after the action. Same shape as `rho`.

    """
    k = len(rho)
    return p * rho + (1 - p) * np.ones(k) / k


# ====Functions related to distance measures==== #


def fidelity(rho, mu):
    """Calculate fidelity of two states given in the same graph state basis.

    This is a special case of the general definition of the fidelity:
    F(rho, mu) = (tr(sqrt(sqrt(rho), mu, sqrt(rho))))**2
    Note that the term "fidelity" has been used ambiguously in quantum
    information theory, either referring to F or sqrt(F). F as defined here is
    the square of the fidelity as defined in Nielsen and Chuang.

    sqrt(1 - F(rho, mu)) is a distance measure.
    (1 - sqrt(F(rho, mu))) is a distance measure.

    Parameters
    ----------
    rho, mu : np.ndarray
        Diagonal entries of quantum states given in the same graph state basis.

    Returns
    -------
    scalar
        The fidelity F.

    """
    a = np.sqrt(rho)
    b = np.sqrt(mu)
    return np.dot(a, b) ** 2


def fid_alternative(rho, mu):
    """Alternative fidelity function.

    Calculates sqrt(F) instead of F (as defined in the `fidelity` fucntion).

    Parameters
    ----------
    rho, mu : np.ndarray
        Diagonal entries of quantum states given in the same graph state basis.

    Returns
    -------
    scalar
        sqrt(F)

    """
    a = np.sqrt(rho)
    b = np.sqrt(mu)
    return np.dot(a, b)


def trace_distance(rho, mu):
    """Calculate the trace distance between to states in the same graph state basis.

    Parameters
    ----------
    rho, mu : np.ndarray
        Diagonal entries of quantum states given in the same graph state basis.

    Returns
    -------
    scalar
        The trace distance.

    """
    sigma = np.abs(rho - mu)
    return 1 / 2 * np.sum(sigma)


# ====Auxiliary Functions==== #


def normalize(rho):
    """Normalize the state to trace = 1.

    Also catches numerical phenomena with entries < 0.

    Parameters
    ----------
    rho : np.ndarray
        The state to be normalized.

    Returns
    -------
    np.ndarray
        The normalized state. Same shape as `rho`.

    """
    if np.any(rho < 0):
        rho = np.copy(rho)
        rho[rho < 0] = 0
    return rho / np.sum(rho)


def local_complementation(n, graph):
    """Return the new graph after local complementation.

    Careful: Subsets are just copied so the coloring is not updated!

    Parameters
    ----------
    n : int
        Local complementation around the `n`-th vertex.
    graph : Graph
        The original graph.

    Returns
    -------
    Graph
        The graph after local complementation.

    """
    # note that Graph is only a simple graph!
    # crude implementation - surely this can be done better.

    # get neighbourhood of n
    Nn = []
    for i in range(graph.N):
        if graph.adj[i, n]:
            Nn += [i]
    new_adjmatrix = np.copy(graph.adj)
    for i in Nn:
        for k in Nn:
            if k == i:
                continue
            new_adjmatrix[i, k] = (graph.adj[i, k] + 1) % 2
    # get new edges from adjmatrix
    new_edges = []
    for i in range(graph.N):
        for k in range(i + 1, graph.N):
            if new_adjmatrix[i, k]:
                new_edges += [(i, k)]
    return Graph(
        N=graph.N, E=new_edges, sets=graph.sets
    )  # just copies sets without thinking


# ====EPP functions for two-colorable states==== #


@lru_cache(maxsize=None)  # this will be getting called a lot with the same input
def _mask_a(j, graph):
    """Spread a bit string on set a to the whole bitstring.

    Takes an int representing a bit string of length size of set a
    and spreads it to a length graph.N bit string with the bits set at the
    correct places.

    Example: graph.N = 4, graph.a = [0, 2], j=3 (bitstring "11")
             will return 10 (bitstring "1010")

    Example: graph.N = 4, graph.a = [0, 2], j=1 (bitstring "01")
             will return 2 (bitstring "0010")

    Parameters
    ----------
    j : int
        Representing a bit string of length len(graph.a).
    graph : Graph
        The graph containing information about the set.

    Returns
    -------
    int
        Representing a bit string of length graph.N, i.e. `j` spread out over
        the appropriate positions in the bit string.

    """
    m = ["0"] * graph.N
    short_string = format(j, "0" + str(len(graph.a)) + "b")
    for bit, idx in zip(short_string, graph.a):
        m[idx] = bit
    long_string = "".join(m)
    return int(long_string, base=2)


@lru_cache(maxsize=None)  # this will be getting called a lot with the same input
def _mask_b(j, graph):
    """Spread a bit string on set b to the whole bitstring.

    Takes an int representing a bit string of length size of set b
    and spreads it to a length graph.N bit string with the bits set at the
    correct places.

    Example: graph.N = 4, graph.b = [1, 3], j=3 (bitstring "11")
             will return 10 (bitstring "0101")

    Example: graph.N = 4, graph.b = [1, 3], j=1 (bitstring "01")
             will return 2 (bitstring "0001")

    Parameters
    ----------
    j : int
        Representing a bit string of length len(graph.b).
    graph : Graph
        The graph containing information about the set.

    Returns
    -------
    int
        Representing a bit string of length graph.N, i.e. `j` spread out over
        the appropriate positions in the bit string.

    """
    m = ["0"] * graph.N
    short_string = format(j, "0" + str(len(graph.b)) + "b")
    for bit, idx in zip(short_string, graph.b):
        m[idx] = bit
    long_string = "".join(m)
    return int(long_string, base=2)


# Note: np.fromfunction does not help with speeding up p1p2, but cythonizing does!
def p1(rho, graph):
    """Perform protocol P1 of the ADB protocol for two-colorable graph states.

    Implements equation (17) of Phys. Rev. A 71, 012319 (2005)
    Preprint: https://arxiv.org/abs/quant-ph/0405045

    Comment on the implementation:
    The integers used here correspond to the bit strings in the publication
    as follows:
    i ~ γ_A,γ_B
    j ~ 0,γ_B
    k ~ ν_B
    m ~ 0,ν_B
    and therefore:
    i^j     ~ γ_A,0
    (i^j)^m ~ γ_A,ν_B
    i^m     ~ γ_A,(γ_B ⊕ ν_B)
    So the loop over k iterates over all possible ν_B. While equation (17)
    suggests another loop over μ_B, there is only one μ_B = (γ_B ⊕ ν_B) that
    fulfils the specified condition ν_B ⊕ μ_B = γ_B so another nested loop is
    not necessary.

    Parameters
    ----------
    rho : np.ndarray
        Diagonal entries of a density matrix in the graph state basis.
        Two copies of this state will be used to perform the protocol.
    graph : Graph
        Graph of the target graph state to be purified. rho is given in this
        graph state basis. Must contain coloring information of the
        two-colorable graph state.

    Returns
    -------
    np.ndarray
        The output state of the protocol, assuming the purification step was
        successful.

    """
    mu = np.zeros(len(rho))
    for i in range(2 ** graph.N):
        j = i & (_mask_b((1 << len(graph.b)) - 1, graph))
        for k in range(2 ** len(graph.b)):
            m = _mask_b(k, graph)
            mu[i] += rho[(i ^ j) ^ m] * rho[i ^ m]
    mu = normalize(mu)
    return mu


def p2(rho, graph):
    """Perform protocol P2 of the ADB protocol for two-colorable graph states.

    Implements equation (19) of Phys. Rev. A 71, 012319 (2005)
    Preprint: https://arxiv.org/abs/quant-ph/0405045

    See docstring of p1 for comment on the iterations and bit strings.

    Parameters
    ----------
    rho : np.ndarray
        Diagonal entries of a density matrix in the graph state basis.
        Two copies of this state will be used to perform the protocol.
    graph : Graph
        Graph of the target graph state to be purified. rho is given in this
        graph state basis. Must contain coloring information of the
        two-colorable graph state.

    Returns
    -------
    np.ndarray
        The output state of the protocol, assuming the purification step was
        successful.

    """
    mu = np.zeros(len(rho))
    for i in range(2 ** graph.N):
        j = i & (_mask_a((1 << len(graph.a)) - 1, graph))
        for k in range(2 ** len(graph.a)):
            m = _mask_a(k, graph)
            mu[i] += rho[(i ^ j) ^ m] * rho[i ^ m]
    mu = normalize(mu)
    return mu


def p1_var(rho, sigma, graph):
    """P1 but with two different input states instead of two copies of the same.

    Parameters
    ----------
    rho : np.ndarray
        Diagonal entries of a density matrix in the graph state basis.
        First input state.
    sigma : np.ndarray
        Diagonal entries of a density matrix in the graph state basis.
        Second input state.
    graph : Graph
        Graph of the target graph state to be purified. rho and mu are
        given in this graph state basis. Must contain coloring information
        of the two-colorable graph state.

    Returns
    -------
    np.ndarray
        The output state of the protocol, assuming the purification step was
        successful.

    """
    mu = np.zeros(len(rho))
    for i in range(2 ** graph.N):
        j = i & (_mask_b((1 << len(graph.b)) - 1, graph))
        for k in range(2 ** len(graph.b)):
            m = _mask_b(k, graph)
            mu[i] += rho[(i ^ j) ^ m] * sigma[i ^ m]
    mu = normalize(mu)
    return mu


def p2_var(rho, sigma, graph):
    """P2 but with two different input states instead of two copies of the same.

    Parameters
    ----------
    rho : np.ndarray
        Diagonal entries of a density matrix in the graph state basis.
        First input state.
    sigma : np.ndarray
        Diagonal entries of a density matrix in the graph state basis.
        Second input state.
    graph : Graph
        Graph of the target graph state to be purified. rho and mu are
        given in this graph state basis. Must contain coloring information
        of the two-colorable graph state.

    Returns
    -------
    np.ndarray
        The output state of the protocol, assuming the purification step was
        successful.

    """
    mu = np.zeros(len(rho))
    for i in range(2 ** graph.N):
        j = i & (_mask_a((1 << len(graph.a)) - 1, graph))
        for k in range(2 ** len(graph.a)):
            m = _mask_a(k, graph)
            mu[i] += rho[(i ^ j) ^ m] * sigma[i ^ m]
    mu = normalize(mu)
    return mu


# ====EPP functions for arbitrary graph states==== #


@lru_cache(maxsize=None)
def _mask_k(j, graph, subset):
    """Spread a bit string on a subset to the whole bitstring.

    Takes an int representing a bit string of length len(myset)
    and spreads it to a length graph.N bit string with the bits set at the
    correct places.

    Example: graph.N = 4, myset = (0, 2), j=3 (bitstring "11")
             will return 10 (bitstring "1010")

    Example: graph.N = 4, myset = (0, 2), j=1 (bitstring "01")
             will return 2 (bitstring "0010")

    Parameters
    ----------
    j : int
        Representing a bit string of length len(subset).
    graph : Graph
        The graph, basically just here for graph.N
    subset : tuple of ints
        A subset of vertices of the graph. Ideally use a tuple not a list to
        allow caching to work.

    Returns
    -------
    int
        Representing a bit string of length graph.N, i.e. `j` spread out over
        the appropriate positions in the bit string.

    """
    m = ["0"] * graph.N
    short_string = format(j, "0" + str(len(subset)) + "b")
    for bit, idx in zip(short_string, subset):
        m[idx] = bit
    long_string = "".join(m)
    return int(long_string, base=2)


def pk(rho, sigma, graph1, graph2, subset):
    """Perform sub-protocol P_k.

    A sub-protocol of the entanglement purification protocol for all graph
    states. Implements equation (8) of Phys. Rev. A 74, 052316 (2006)
    Preprint: https://arxiv.org/abs/quant-ph/0606090

    See docstring of p1 for comment on the iterations and bit strings.

    Parameters
    ----------
    rho : np.ndarray
        Diagonal entries of a density matrix in the graph state basis
        corresponding to `graph1`. Main input state.
    sigma : np.ndarray
        Diagonal entries of a density matrix in the graph state basis
        corresponding to `graph2`. Auxiliary input state. Make sure it
        has the same number of qubits as `rho` (expand with unconnected
        vertices if needed).
    graph1 : Graph
        The main graph of the protocol.
    graph2 : Graph
        The auxiliary graph for the k-th subset for the k-th sub-protocol P_k.
        Make sure it has the same number of vertices as `graph1`
        (expand with unconnected vertices if needed).
    subset : tuple of ints
        A subset of vertices, corresponding to the k-th subset.

    Returns
    -------
    np.ndarray
        The output state of the protocol, assuming the purification step was
        successful.

    """
    mu = np.zeros(len(rho))
    other_set = tuple(i for i in range(graph1.N) if i not in subset)
    for i in range(2 ** graph1.N):
        j = i & (_mask_k((1 << len(other_set)) - 1, graph1, other_set))
        for k in range(2 ** len(other_set)):
            m = _mask_k(k, graph1, other_set)
            mu[i] += rho[(i ^ j) ^ m] * sigma[i ^ m]
    mu = normalize(mu)
    return mu



"""Applying the multipartite entanglement purification protocol."""

import os, sys

sys.path.insert(0, os.path.abspath(".."))
import graphepp as gg
import numpy as np

# Example 1: apply P1 and P2 to a noisy GHZ state
print("Example 1: apply P1 and P2 to a noisy GHZ state")

# graph state version of the GHZ state
ghz_graph = gg.Graph(N=3, E=((0, 1), (0, 2)), sets=[[0], [1, 2]])
# the sets indicate the coloring of this two-colorable graph state
# note that vertices are counting from 0...N-1

rho = np.zeros(2 ** ghz_graph.N, dtype=np.float)
rho[0] = 1.0  # this means the state is the perfect graph state
perfect_state = np.copy(rho)
# start with an initally noise state
rho = gg.wnoise_all(rho=rho, p=0.95, graph=ghz_graph)
print(f"Fidelity before: {gg.fidelity(rho, perfect_state):.5f}")
# now apply P1 once then P2 once
rho = gg.p1(rho=rho, graph=ghz_graph)
rho = gg.p2(rho=rho, graph=ghz_graph)
print(f"Fidelity after: {gg.fidelity(rho, perfect_state):.5f}")


# Example 2: Show that a sufficiently high-fidelity state will converge to the
#            the perfect graph state. Here: Linear cluster state of 5 qubits
print("=================")
print("Example 2: Linear cluster state convergence")

linear_cluster_graph = gg.Graph(
    N=5, E=((0, 1), (1, 2), (2, 3), (3, 4)), sets=[[0, 2, 4], [1, 3]]
)
rho = np.zeros(2 ** linear_cluster_graph.N, dtype=np.float)
rho[0] = 1.0  # the perfect graph state
# first consider a slightly noisy state
mu = gg.wnoise_all(rho=rho, p=0.95, graph=linear_cluster_graph)
print("a) Purification works for sufficiently high-fidelity state.")
print(f"Slightly noisy state initial fidelity: {gg.fidelity(mu, rho):.5f}")
# 100 purification steps should be more than enough to get to the fixed point
# within numerical precision
for i in range(100):
    mu = gg.p1(rho=mu, graph=linear_cluster_graph)
    mu = gg.p2(rho=mu, graph=linear_cluster_graph)
# if initial fidelity is high enough this should converge to the perfect state
print(f"Fidelity after 100 purification steps: {gg.fidelity(mu, rho):.5f}")
print("---------")
mu = gg.wnoise_all(rho=rho, p=0.5, graph=linear_cluster_graph)
print("b) Purification fails for initial states that are too noisy.")
print(f"Very noisy state initial fidelity: {gg.fidelity(mu, rho):.5f}")
for i in range(100):
    mu = gg.p1(rho=mu, graph=linear_cluster_graph)
    mu = gg.p2(rho=mu, graph=linear_cluster_graph)
print(f"Fidelity after 100 purification steps: {gg.fidelity(mu, rho):.5f}")

print("=================")
print("Example 3: GHZ Fidelity change during purification")
ghz_graph = gg.Graph(N=3, E=((0, 1), (0, 2)), sets=[[0], [1, 2]])


rho = np.zeros(2 ** ghz_graph.N, dtype=np.float)
rho[0] = 1.0
perfect_state = np.copy(rho)
# start with an initally noise state
rho = gg.wnoise_all(rho=rho, p=0.8, graph=ghz_graph)
fidelities = [gg.fidelity(rho, perfect_state)]
for i in range(7):
    rho = gg.p1(rho=rho, graph=ghz_graph)
    fidelities += [gg.fidelity(rho, perfect_state)]
    rho = gg.p2(rho=rho, graph=ghz_graph)
    fidelities += [gg.fidelity(rho, perfect_state)]
print(f"Fidelity change over 14 purification steps: {fidelities}")
print("If pyplot is available, will show a plot.")

try:
    import matplotlib.pyplot as plt

    plotting_possible = True
except ImportError:
    plotting_possible = False

if plotting_possible:
    plt.scatter(np.arange(len(fidelities)), fidelities)
    plt.ylim(0.6, 1.01)
    plt.xlabel("Purification steps")
    plt.ylabel("Fidelity")
    plt.grid()
    plt.show()



from setuptools import setup

with open("README.md", "r", encoding="utf-8") as readme:
    long_description = readme.read()

setup(
    name="graphepp",
    description="A collection of functions for multipartite entanglement purification protocols (EPP) on noisy graph states",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Julius Wallnöfer",
    url="https://github.com/jwallnoefer/graphepp",
    license="MIT",
    license_files=["LICENSE"],
    install_requires=["numpy>=1.19.2"],
    package_dir={"": "src"},
    packages=["graphepp"],
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Physics",
        "Intended Audience :: Science/Research",
    ],
    keywords=[
        "quantum information",
        "quantum information theory",
        "graph",
        "graph state",
        "entanglement",
        "entanglement purification",
        "entanglement purification protocol",
        "EPP",
        "entanglement distillation",
        "multipartite entanglement",
    ],
    python_requires=">=3.8.5",
    use_scm_version={
        "write_to": "src/graphepp/version.py",
        "local_scheme": "no-local-version",
    },
    setup_requires=["setuptools_scm"],
)
